/****************************************
 * File:    huffman.cpp                 *
 * Author:  Konrad Rudolph              *
 * Created: 20/05/2004                  *
 * Changed: 25/05/2004                  *
 *                                      *
 * Special thanks to Achim Neubauer     *
 ****************************************/
#include <windows.h>
#include "Huffman.h"

/////////////////////////////////////////////
// public interface

Huffman::Huffman() :
                    _First(0),
                    _Root(0),
                    _State(NoInit)
{
    // assign file header ID + version
    _HeaderID[0] = 'H';
    _HeaderID[1] = 'U';
    _HeaderID[2] = 'F';
    _HeaderID[3] = 1;
    
    // initialize bit table
    _BitTable[0] = 1;
    
    for(int i = 1; i < 32; i++)
        _BitTable[i] = _BitTable[i - 1] << 1;
}

Huffman::~Huffman()
{
    Clear();
}

int Huffman::GetEncodeLength(const char* Input, int Len)
{
    // too small for efficient compression - leave unharmed
    if (Len < 256 * sizeof(int) + 1) {
        _State = NoCompress;
        return Len;
    }
    
    Create(Input, Len);
    
    _BuffLen = 0;
    
    // add lengths of all chars to obtain total length
    for (int i = 0; i < Len; i++)
        _BuffLen += _Table[Input[i]].Len;
    
    // conversion bits -> bytes
    _BuffLen = ((_BuffLen - 1) >> 3) + 1;
    _State = Compress;
    
    // add length of header to result
    return _BuffLen + 5 + 257 * sizeof(int);
}

int Huffman::GetDecodeLength(const char* Input, int Len)
{
    for (int i = 0; i < 4; i++)
        if (Input[i] != _HeaderID[i]) {
            // no compressed input data - return unchanged
            _State = NoDecompress;
            return Len;
        }
    
    Read(Input, 4);
    
    // read compressed data length
    const int TmpLen = *reinterpret_cast<const int*>(&Input[4 + 256 * sizeof(int)]);
    _BuffLen = TmpLen;
    _State = Decompress;
    
    return _BuffLen;
}

bool Huffman::Encode(char const* Input, char* Output, int Len)
{
    int     Offset  = 0,
            Cell    = 0;
    char*   Buff;
    
    if (_State == NoCompress) {
        // return input string
        memcpy(Output, Input, Len + 1);
        return true;
    }
    
    if (_State != Compress)
        return false;
    
    // allocate memory for temporary data
    Buff = new char[_BuffLen];
    
    // reset value
    Buff[0] = 0;
    
    // for all chars ...
    for (int i = 0; i < Len; i++) {
        // for every bit in char's code ...
        for(int Bit = _Table[Input[i]].Len - 1; Bit >= 0; Bit--) {
            // if current bit is set then set output bit
            if (_Table[Input[i]].Bits & _BitTable[Bit])
                Buff[Cell] |= _BitTable[Offset];
            
            Offset++;
            
            if (Offset == 8) {
                Offset = 0;
                Buff[++Cell] = 0;
            } // if
        } // for
    } // for
    
    char*   Head;
    int     HeadLen;
    
    Serialize(&Head, HeadLen);
    
    // concatenate all this data
    
    // write header ID
    memcpy(Output, _HeaderID, 4);
    
    // write header
    memcpy(&Output[4], Head, HeadLen);
    
    // write length
    memcpy(&Output[4 + HeadLen], &Len, sizeof(int));
    
    // write data
    memcpy(&Output[4 + HeadLen + sizeof(int)], Buff, _BuffLen);
    
    Output[4 + HeadLen + sizeof(int) + _BuffLen] = 0;
    
    // destroy temporary objects
    delete [] Buff;
    delete [] Head;
    
    return true;
}

bool Huffman::Decode(char const* Input, char* Output, int Len)
{
    if (_State == NoDecompress) {
        memcpy(Output, Input, Len + 1);
        return true;
    }
    
    if (_State != Decompress)
        return false;
    
    // starting cell behind header data
    int     Cell    = 4 + 257 * sizeof(int);
    int     Offset  = 0;
    int     i;
    
    // for each char
    for (i = 0; i < _BuffLen; i++) {
        Node*   Iter    = _Root;
        
        // follow char's byte path in Huffman tree
        while (Iter->Left != 0) {
            if (Input[Cell] & _BitTable[Offset++])
                Iter = Iter->Right;
            else
                Iter = Iter->Left;
            
            if (Offset == 8) {
                Offset = 0;
                Cell++;
            } // if
        } // while
        
        Output[i] = Iter->Char;            
    } // for
    
    Output[i] = 0; // zero termination
    return true;
}

/////////////////////////////////////////////
// implementation

// list functions

void Huffman::Append(Node* Item)
{
    Item->Next = _First;
    _First = Item;
}

Node* Huffman::PopLowest(void)
{
    Node*   Iter    = _First;
    Node*   Last    = 0;
    Node*   MinItem = 0;
    long    Min     = 0xffffffL;
    
    // loop through all values
    while (Iter) {
        if (Iter->Freq < Min) {
            Min = Iter->Freq;
            MinItem = Last;
        }
        
        Last = Iter;
        Iter = Iter->Next;
    }
    
    // remove minimum from list
    if (MinItem == 0) {
        Last = _First;
        _First = _First->Next;
    }
    else {
        Last = MinItem->Next;
        MinItem->Next = Last->Next;
    }
    
    return Last;
}

// dictionary functions

void Huffman::Create(char const* Text, const int Len)
{
    Init();
    
    // count characters' frequencies
    for (int i = 0; i < Len; i++)
        _Freqs[Text[i]]++;
    
    _Root = CreateHuffmanTree();
    
    ByteCode NewCode = {0, 0};
    CreateTransitionTable(_Root, NewCode);
}

void Huffman::Serialize(char** Result, int& Len)
{
    Len = 256 * sizeof(int);
    *Result = new char[Len];
    
    memcpy(*Result, &_Freqs, 256 * sizeof(int));
}

void Huffman::Read(char const* Code, int Offset)
{
    Init();
    
    memcpy(&_Freqs, &Code[Offset], 256 * sizeof(int));
    
    _Root = CreateHuffmanTree();
}

Node* Huffman::CreateHuffmanTree(void)
{
    // put all nodes into container
    
    for (int i = 0; i < 256; i++) {
        if (_Freqs[i] > 0) {
            Node* NewNode = new Node();
            
            NewNode->Left = 0;
            NewNode->Right = 0;
            NewNode->Char = static_cast<int>(i);
            NewNode->Freq = _Freqs[i];
            
            Append(NewNode);
        }
    }
    
    Node*   Left,
        *   Right;
    
    while (_First) {
        Left = PopLowest();
        Right = PopLowest();
        
        Node* NewNode = new Node();
        NewNode->Freq = Left->Freq + Right->Freq;
        NewNode->Left = Left;
        NewNode->Right = Right;
        
        if (!_First)
            return NewNode;
        
        Append(NewNode);
    }
    
    return 0;
}

void Huffman::CreateTransitionTable(Node* CurrNode, ByteCode Code)
{
    if (CurrNode->Left == 0)
        _Table[static_cast<int>(CurrNode->Char)] = Code;
    else {
        ByteCode NewCode;
        
        NewCode.Len = Code.Len + 1;
        NewCode.Bits = Code.Bits << 1;
        CreateTransitionTable(CurrNode->Left, NewCode);
        
        NewCode.Bits |= 1;
        CreateTransitionTable(CurrNode->Right, NewCode);
    }
}

void Huffman::Init(void)
{
    Clear();
    _First = 0;
    
    for (int i = 0; i < 256; i++)
        _Freqs[i] = 0;
}

void Huffman::Clear(void)
{
    while (_First) {
        Node* tmp = _First;
        _First = _First->Next;
        delete tmp;
    }
}