/****************************************
 * File:       Huffman.h                *
 * Author:     Konrad Rudolph           *
 * Created:    20/05/2004               *
 * Changed:    25/05/2004               *
 ****************************************/

#if (!defined(_huffman_klmr_15052004_))
#define _huffman_klmr_15052004_ 1

// data node
struct Node {
    char    Char;                   // character value
    int     Freq;                   // character's frequency
    Node*   Left;                   // left child
    Node*   Right;                  // right child
    Node*   Next;                   // next sibling in list
};

// byte code structure
struct ByteCode {
    int     Bits;                   // bit value
    int     Len;                    // bit length
};

// Huffman code wrapper
class Huffman
{
    public:
        Huffman();                  // c'tor
        ~Huffman();                 // d'tor
        
        int GetEncodeLength(        // returns length of the encoded text
            char const* Input,      // [IN] input text
            int Len                 // [IN] input length
        );
        
        int GetDecodeLength(        // returns length of the decoded text
            char const* Input,      // [IN] input text
            int Len                 // [IN] input length
        );
        
        bool Encode(                // Encode data - returns success
            char const* Input,      // [IN] plain data
            char* Output,           // [OUT] encoded data
            int Len                 // [IN] data length
        );
        bool Decode(                // Decode data - returns success
            char const* Input,      // [IN] encoded data
            char* Output,           // [OUT] plain data
            int Len                 // [IN] encoded data length
        );

    private:
        // different action states of class
        enum ActionState {
            NoInit, Compress, Decompress, NoCompress, NoDecompress
        };
        
        // list members
        Node*       _First;         // first list item
        
        // dictionary members
        Node*       _Root;          // tree root item
        ByteCode    _Table[256];    // byte codes
        int         _Freqs[256];    // character frequencies
        
        // huffman members
        char        _HeaderID[4];   // encoded data file header
        ActionState _State;         // current action state of class
        int         _BuffLen;       // buffer lenght
        
        int         _BitTable[32];  // 2's powers for speed
        
        // list functions
        void Append(                // appends an item to the list
            Node* Item              // [IN] item to append
        );
        Node* PopLowest(void);      // - returns and removes lowest item
        
        // dictionary functions
        void Create(                // create the dictionary
            char const* Text,       // [IN] reference text
            const int Len           // [IN] text length
        );
        void Serialize(             // serialize dictionary
            char** Result,          // [OUT] data
            int& Len                // [OUT] data length
        );
        void Read(                  // read serialized dict
            char const* Code,       // [IN] code data
            int Offset              // [IN] initial offset
        );
        Node* CreateHuffmanTree(void); // creates huffman tree - returns root
        void CreateTransitionTable( // creates translation table recursively
            Node* Node,             // [IN] current node
            ByteCode Code           // [IN] current byte code
        );
        
        void Init(void);            // initialize everything
        void Clear(void);           // clear everything, free memory
};

#endif // (!defined _huffman_klmr_15052004_)